/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2007, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Parameter settings for the busy tone detection and calling identity delivery analysis
 * 
 * \author kevin.chen
 *
 * \Copyright (C) 2012 OpenVox Communication Co. Ltd
 *
 *
 * Rev history
 *
 * Rev 0.10 added caller id analysis parameter settings
 *
 */

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 328209 $")

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#include "asterisk/module.h"
#include "asterisk/cli.h"
#include "asterisk/config.h"
#include "asterisk/pbx.h"

/* Compatible with earlier versions */
#ifndef AST_MODULE_INFO
# define AST_VERSION_1_2
#elif !defined(AST_CLI_DEFINE)
# define AST_VERSION_1_4
#endif

#ifndef FALSE
#define FALSE   0
#endif

#ifndef TRUE
#define TRUE    1
#endif

/* Configuration file */
#define OPVXDSP_CONFIG	"opvxdsp.conf"

/* Top-level directory */
static const char *path = "/proc/opvxdsp";

/* Caller id subdirectory */
static const char *cid_subdir = "opencid";

/* The current channel, for parameter settings. Not set by default */
static int g_chan = 0;

/* The busy tone detection subdirectory and parameters information */
static const struct bd_param {
	const char *subdir;	/* Subdirectory */
	const char *variable[10];  /* A variety of parameters */
}bd_params[] = {
        {"tone1", {"busycount", "threshold", "ontime", "offtime", "frequency1", "frequency2", NULL}},
        {"tone2", {"busycount", "threshold", "ontime", "offtime", "frequency1", "frequency2", NULL}},
        {"tone3", {"busycount", "threshold", "ontime", "offtime", "frequency1", "frequency2", NULL}},
        {"silent_detect", {"detect_tx", "detect_rx", "length", "threshold", NULL}},	
};

/* Caller id parameters information */
static const struct cid_param {
        int writable;   /* Allow write operation */
        const char *variable;
}cid_params[] = {
        {FALSE, "cidstart"},
        {FALSE, "detect_signal"},
        {FALSE, "last_signal"},
        {TRUE, "disable"},
};

static int is_valid_chan(int chan)
{
	DIR *dir;
	char tmp[256];

	sprintf(tmp, "%s/%d", path, chan);

	dir = opendir(tmp);
	if (dir != NULL) {
		/* It is an effective channel */
		closedir(dir);
		return 1;
	} else {
		return 0;
	}
}

static char *get_param_string(const char *subdir, const char *filename)
{
	FILE *f;
	char file[256];
	char buf[24];
	char *p = buf;

	sprintf(file, "%s/%d/%s/%s", path, g_chan, subdir, filename);

	if ((f = fopen(file, "r")) != NULL) {
		memset(buf, 0, sizeof(buf));
		fgets(buf, sizeof(buf), f);
		fclose(f);
		return p;
	} else {
		/* Read file failed, return an NULL value */
		return NULL;
	}
}

static int set_param_value(const char *subdir, const char *filename, int val)
{
	int res = 0;
	char file[256];
	char buf[24];
	FILE *f;

	if (g_chan == 0) {
		/* Not set the current channel, return */
		return res;
	}

	sprintf(file, "%s/%d/%s/%s", path, g_chan, subdir, filename);

	if ((f = fopen(file, "w+")) != NULL) {
		sprintf(buf, "%d", val);
		fwrite(buf, strlen(buf), 1, f);
		fclose(f);

		if ((f = fopen(file, "r")) != NULL) {
			memset(buf, 0, sizeof(buf));
			fgets(buf, sizeof(buf), f);
			fclose(f);
			/* Whether the consistency of modification */
			if (val == atoi(buf)) {
				res = 1;
			}
		}
	}

	return res;
}

static void show_all_channel_number(int fd)
{
	DIR *dir;
	struct dirent *ent;

	ast_cli(fd, "Fxo channel numbers:\n");

	dir = opendir(path);
	if (dir) {
		/* Search all FXO channel */
		while ((ent = readdir(dir))) {
			if (ent->d_type == DT_DIR && ent->d_name[0] != '.') {
				ast_cli(fd, "%s ", ent->d_name);
			}
		}

		ast_cli(fd, "\n");
		closedir(dir);
	}
}

static void busydetect_show_channel_param(int fd)
{
	int i, j;
	char tmp[256];

	ast_cli(fd, "Channel number: %d\n", g_chan);
	
	for (i = 0; i < sizeof(bd_params) / sizeof(bd_params[0]); i++) {
		sprintf(tmp, "%s:", bd_params[i].subdir);
		for (j = 0; bd_params[i].variable[j] != NULL; j++) {
			sprintf(tmp, "%s %s=%s,", tmp, bd_params[i].variable[j], 
				get_param_string(bd_params[i].subdir, bd_params[i].variable[j]));
		}
		/* Replace the line breaks */
		tmp[strlen(tmp) - 1] = '\n';
		
		/* Output information */
		ast_cli(fd, tmp);
	}
}

static void busydetect_set_channel_param(int fd, const char *subdir, const char *filename, int val)
{
	if (set_param_value(subdir, filename, val)) {
		ast_cli(fd, "Set %d channel parameter success\n", g_chan);
	} else {
		if (g_chan == 0) {
			ast_cli(fd, "Set channel parameter failed\n");
			ast_cli(fd, "Need to set the current channel\n"
				"perform: busydetect show [fxo channel number]\n");	
		} else {
			ast_cli(fd, "Set %d channel parameter failed\n", g_chan);
		}
	}
}

static void callerid_show_channel_param(int fd)
{
        int i;
        char tmp[256];

        ast_cli(fd, "Channel number: %d\n", g_chan);

        sprintf(tmp, "%s=%s,", cid_params[0].variable, get_param_string(cid_subdir, cid_params[0].variable));
        for (i = 1; i < sizeof(cid_params) / sizeof(cid_params[0]); i++) {
                sprintf(tmp, "%s %s=%s,", tmp, cid_params[i].variable, get_param_string(cid_subdir, cid_params[i].variable));
        }

        /* Replace the line breaks */
        tmp[strlen(tmp) - 1] = '\n';

        /* Output information */
        ast_cli(fd, tmp);
}

static void callerid_set_channel_param(int fd, const char *filename, int val)
{
        if (set_param_value(cid_subdir, filename, val)) {
                ast_cli(fd, "Set %d channel parameter success\n", g_chan);
        } else {
                if (g_chan == 0) {
                        ast_cli(fd, "Set channel parameter failed\n");
                        ast_cli(fd, "Need to set the current channel\n"
                                "perform: callerid show [fxo channel number]\n");
                } else {
                        ast_cli(fd, "Set %d channel parameter failed\n", g_chan);
                }
        }
}

#if defined(AST_VERSION_1_2) || defined(AST_VERSION_1_4)
static int busydetect_show(int fd, int argc, char **argv)
{
	int res = RESULT_SHOWUSAGE;
	int chan;

	if (argc == 2) {
		show_all_channel_number(fd);
		res = RESULT_SUCCESS;
	} else if (argc == 3) {
		chan = atoi(argv[2]);
		if (chan > 0 && is_valid_chan(chan)) {
			/* Set the current channel */
			g_chan = chan;
			busydetect_show_channel_param(fd);
			res = RESULT_SUCCESS; 
		}
	}

	return res;
}

static int busydetect_set_tone(int fd, int argc, char **argv)
{
	int val;
	
	if (argc != 5) {
		return RESULT_SHOWUSAGE;
	}

	val = atoi(argv[4]);
	if (val < 0 || (!strncmp(argv[3], "frequency", 9) && (val < 300 || val > 700))) {
		return RESULT_SHOWUSAGE; 
	}
		
	busydetect_set_channel_param(fd, argv[2], argv[3], val);
	return RESULT_SUCCESS; 
}

static int busydetect_set_silent(int fd, int argc, char **argv)
{
	int val;

	if (argc != 5) {
		return RESULT_SHOWUSAGE;
	}

	val = atoi(argv[4]);
	if (val < 0) {
		return RESULT_SHOWUSAGE; 
	}

	busydetect_set_channel_param(fd, argv[2], argv[3], val);
	return RESULT_SUCCESS; 
}
#else
static char *busydetect_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char *res = CLI_SHOWUSAGE;
	int chan;

	switch (cmd) {
	case CLI_INIT:
		e->command = "busydetect show";
		e->usage = 
			"Usage: busydetect show [fxo channel number]\n"
			"	Example:\n"
			"	busydetect show\n"
			"	busydetect show 1\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 2) {
		show_all_channel_number(a->fd);
		res = CLI_SUCCESS;
	} else if (a->argc == 3) {
		chan = atoi(a->argv[2]);
		if (chan > 0 && is_valid_chan(chan)) {
			/* Set the current channel */
			g_chan = chan;
			busydetect_show_channel_param(a->fd);
			res = CLI_SUCCESS; 
		}
	}

	return res;
}

static char *busydetect_set_tone(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int val;

	switch (cmd) {
	case CLI_INIT:
		e->command = "busydetect set {tone1|tone2|tone3} {busycount|threshold|ontime|offtime|frequency1|frequency2}";
		e->usage = 
			"Usage: busydetect set {tone1|tone2|tone3} {busycount|threshold|ontime|offtime|frequency1|frequency2}\n"
			"	Valid values greater than or equal to zero\n"
			"	ontime, offtime: unit millisecond\n"
			"	frequency1, frequency2: valid values are in the range 300-700HZ\n"
			"	Example:\n"
			"	busydetect set tone1 busycount 3\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 5) {
		return CLI_SHOWUSAGE;
	}

	val = atoi(a->argv[4]);
	if (val < 0 || (!strncmp(a->argv[3], "frequency", 9) && (val < 300 || val > 700))) {
		return CLI_SHOWUSAGE;
	}

	busydetect_set_channel_param(a->fd, a->argv[2], a->argv[3], val);
	return CLI_SUCCESS; 
}

static char *busydetect_set_silent(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int val;

	switch (cmd) {
	case CLI_INIT:
		e->command = "busydetect set silent_detect {detect_tx|detect_rx|length|threshold}";
		e->usage = 
			"Usage: busydetect set silent_detect {detect_tx|detect_rx|length|threshold}\n"
			"	Valid values greater than or equal to zero\n"
			"	detect_tx, detect_rx: 1 support, 0 not support\n"
			"	length: unit second\n"
			"	Example:\n"
			"	busydetect set silent_detect length 30\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 5) {
		return CLI_SHOWUSAGE;
	}

	val = atoi(a->argv[4]);
	if (val < 0) {
		return CLI_SHOWUSAGE; 
	}
	
	busydetect_set_channel_param(a->fd, a->argv[2], a->argv[3], val);
	return CLI_SUCCESS; 
}
#endif

#if defined(AST_VERSION_1_2) || defined(AST_VERSION_1_4)
static char usage_show[] = 
			"Usage: busydetect show [fxo channel number]\n"
			"	Example:\n"
			"	busydetect show\n"
			"	busydetect show 1\n";

static char usage_set_tone[] = 
			"Usage: busydetect set {tone1|tone2|tone3} {busycount|threshold|ontime|offtime|frequency1|frequency2}\n"
			"	Valid values greater than or equal to zero\n"
			"	ontime, offtime: unit millisecond\n"
			"	frequency1, frequency2: valid values are in the range 300-700HZ\n"
			"	Example:\n"
			"	busydetect set tone1 busycount 3\n";

static char usage_set_silent[] =
			"Usage: busydetect set silent_detect {detect_tx|detect_rx|length|threshold}\n"
			"	Valid values greater than or equal to zero\n"
			"	detect_tx, detect_rx: 1 support, 0 not support\n"
			"	length: unit second\n"
			"	Example:\n"
			"	busydetect set silent_detect length 30\n";

static struct ast_cli_entry cli_busydetect[] = {
	{{"busydetect", "show", NULL}, busydetect_show,	"Displays the busy tone detection parameters", usage_show},

	/* tone1 */
	{{"busydetect", "set", "tone1", "busycount", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone1", "threshold", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone1", "ontime", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone1", "offtime", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone1", "frequency1", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone1", "frequency2", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},

	/* tone2 */
	{{"busydetect", "set", "tone2", "busycount", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone2", "threshold", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone2", "ontime", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone2", "offtime", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone2", "frequency1", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone2", "frequency2", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},

	/* tone3 */
	{{"busydetect", "set", "tone3", "busycount", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone3", "threshold", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone3", "ontime", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone3", "offtime", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone3", "frequency1", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},
	{{"busydetect", "set", "tone3", "frequency2", NULL},
		busydetect_set_tone, "Set the tone parameters of the current channel", usage_set_tone},

	/* silent_detect */
	{{"busydetect", "set", "silent_detect", "detect_tx", NULL}, busydetect_set_silent,
		"Set the silent parameters of the current channel", usage_set_silent},
	{{"busydetect", "set", "silent_detect", "detect_rx", NULL}, busydetect_set_silent,
		"Set the silent parameters of the current channel", usage_set_silent},
	{{"busydetect", "set", "silent_detect", "length", NULL}, busydetect_set_silent,
		"Set the silent parameters of the current channel", usage_set_silent},
	{{"busydetect", "set", "silent_detect", "threshold", NULL}, busydetect_set_silent,
		"Set the silent parameters of the current channel", usage_set_silent},
};
#else
static struct ast_cli_entry cli_busydetect[] = {
	AST_CLI_DEFINE(busydetect_show, "Displays the busy tone detection parameters"),
	AST_CLI_DEFINE(busydetect_set_tone, "Set the tone parameters of the current channel"),
	AST_CLI_DEFINE(busydetect_set_silent, "Set the silent parameters of the current channel"),
};
#endif

#if defined(AST_VERSION_1_2) || defined(AST_VERSION_1_4)
static int callerid_show(int fd, int argc, char **argv)
{
	int res = RESULT_SHOWUSAGE;
	int chan;

	if (argc == 2) {
		show_all_channel_number(fd);
		res = RESULT_SUCCESS;
	} else if (argc == 3) {
		chan = atoi(argv[2]);
		if (chan > 0 && is_valid_chan(chan)) {
			/* Set the current channel */
			g_chan = chan;
			callerid_show_channel_param(fd);
			res = RESULT_SUCCESS; 
		}
	}

	return res;
}

static int callerid_set(int fd, int argc, char **argv)
{
	int val;
	
	if (argc != 4) {
		return RESULT_SHOWUSAGE;
	}

	val = atoi(argv[3]) == 0 ? 0 : 1;
	callerid_set_channel_param(fd, argv[2], val);

	return RESULT_SUCCESS; 
}

#else
static char *callerid_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char *res = CLI_SHOWUSAGE;
	int chan;

	switch (cmd) {
	case CLI_INIT:
		e->command = "callerid show";
		e->usage = 
			"Usage: callerid show [fxo channel number]\n"
			"	Example:\n"
			"	callerid show\n"
			"	callerid show 1\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 2) {
		show_all_channel_number(a->fd);
		res = CLI_SUCCESS;
	} else if (a->argc == 3) {
		chan = atoi(a->argv[2]);
		if (chan > 0 && is_valid_chan(chan)) {
			/* Set the current channel */
			g_chan = chan;
			callerid_show_channel_param(a->fd);
			res = CLI_SUCCESS; 
		}
	}

	return res;
}

static char *callerid_set(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int val;

	switch (cmd) {
	case CLI_INIT:
		e->command = "callerid set disable {0|1}";
		e->usage = 
			"Usage: callerid set disable <Boolean>\n"
			"	1 disable caller id analysis, 0 enable\n"
			"	Example:\n"
			"	callerid set disable 0\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}

	val = atoi(a->argv[3]) == 0 ? 0 : 1;
	callerid_set_channel_param(a->fd, a->argv[2], val);

	return CLI_SUCCESS; 
}
#endif

#if defined(AST_VERSION_1_2) || defined(AST_VERSION_1_4)
static char usage_show[] = 
			"Usage: callerid show [fxo channel number]\n"
			"	Example:\n"
			"	callerid show\n"
			"	callerid show 1\n";

static char usage_set[] =
			"Usage: callerid set disable <Boolean>\n"
			"	1 disable caller id analysis, 0 enable\n"
			"	Example:\n"
			"	callerid set disable 0\n";

static struct ast_cli_entry cli_callerid[] = {
	{{"callerid", "show", NULL}, callerid_show, "Displays caller id parameters", usage_show},
	{{"callerid", "set", "disable", "0", NULL}, callerid_set, "Enable the current channel caller id analysis", usage_set},
	{{"callerid", "set", "disable", "1", NULL}, callerid_set, "Disable the current channel caller id analysis", usage_set},
};
#else
static struct ast_cli_entry cli_callerid[] = {
	AST_CLI_DEFINE(callerid_show, "Displays caller id parameters"),
	AST_CLI_DEFINE(callerid_set, "Set disable parameter of the current channel"),
};
#endif

static int load_config(int reload)
{
	int i, oldchan;
	int index;
	struct ast_config *cfg;
	struct ast_variable *var;
	char *cat = NULL;
	const char *p;
        int st, ed;
        char *tok, *q;

#if defined(AST_VERSION_1_2) || defined(AST_VERSION_1_4)
	cfg = ast_config_load(OPVXDSP_CONFIG);
#else
        struct ast_flags config_flags = {reload ? CONFIG_FLAG_FILEUNCHANGED : 0};
        cfg = ast_config_load(OPVXDSP_CONFIG, config_flags);
        if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
                return 0;
        }
#endif
	if (!cfg) {
		ast_log(LOG_ERROR, "Configuration file %s missing.\n", OPVXDSP_CONFIG);
		return -1;
	}

	while ((cat = ast_category_browse(cfg, cat))) {
		index = -1;
		for (i = 0; i < sizeof(bd_params) / sizeof(bd_params[0]); i++) {
			if (!strcasecmp(cat, bd_params[i].subdir)) {
				index = i;
				break;
			}
		}
		
		if (index >= 0 || !strcasecmp(cat, cid_subdir)) {
			p = ast_variable_retrieve(cfg, cat, "channel");
                        if (!p) {
                                continue;
                        }
			/* Have set the channel number, allow modify */
			oldchan = g_chan;

                        tok = strtok((char *)p, ",");
                        while (tok) {
                                st = atoi(tok);
                                q = strchr(tok, '-');
                                if (q) {
                                        ed = atoi(q + 1);
                                } else {
                                        ed = st;
                                }

                                var = ast_variable_browse(cfg, cat);
                                while (var) {
					if (index >= 0) {
						/* parameter settings for the busy tone detection */
						for (i = 0; bd_params[index].variable[i] != NULL; i++) {
							if (!strcasecmp(var->name, bd_params[index].variable[i])) {
								for (g_chan = st; g_chan <= ed; g_chan++) {
									set_param_value(bd_params[index].subdir, var->name, atoi(var->value));	
								}
								break;
							}
						} 
					} else {
						/* parameter settings for calling identity delivery analysis */
						for (i = 0; i < sizeof(cid_params) / sizeof(cid_params[0]); i++) {
							if (cid_params[i].writable &&
							!strcasecmp(var->name, cid_params[i].variable)) {
								for (g_chan = st; g_chan <= ed; g_chan++) {
									set_param_value(cid_subdir, var->name, atoi(var->value));
								}
								break;
							}
						}
					}

                                        var = var->next;
                                } /* while (var) */

                                tok = strtok(NULL, ",");
                        } /* while (tok) */
			g_chan = oldchan;
		} /* if (index >= 0 ... */
	} /* while ((cat = ... */

	return 0;
}

#ifdef AST_VERSION_1_2
int reload(void)
#else
static int reload(void)
#endif
{
        load_config(1);
        return 0;
}

#ifdef AST_VERSION_1_2
int load_module(void)
#else
static int load_module(void)
#endif
{
	load_config(0);
	ast_cli_register_multiple(cli_busydetect, sizeof(cli_busydetect) / sizeof(cli_busydetect[0]));
	ast_cli_register_multiple(cli_callerid, sizeof(cli_callerid) / sizeof(cli_callerid[0]));
	return 0;
}

#ifdef AST_VERSION_1_2
int unload_module(void)
#else
static int unload_module(void)
#endif
{
	ast_cli_unregister_multiple(cli_busydetect, sizeof(cli_busydetect) / sizeof(cli_busydetect[0]));
	ast_cli_unregister_multiple(cli_callerid, sizeof(cli_callerid) / sizeof(cli_callerid[0]));
	return 0;
}

#ifdef AST_VERSION_1_2
char *description(void)
{
	return "Parameter settings for the busy tone detection and calling identity delivery analysis";
}

int usecount(void)
{
	return 1;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
#else
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, 
	"Parameter settings for the busy tone detection and calling identity delivery analysis",
        .load = load_module,
        .unload = unload_module,
        .reload = reload,
        );
#endif
